import rich

from .model import Model
from .view import View


class Control:
    def __init__(self, view: View, model: Model):
        self._view = view
        self._model = model

        modules = self._model.get_ui_modules()

        for _, mod in modules.items():
            try:
                funcs = self._model.get_ui_buttons_functions(mod)

                # Our functions are all decorated functions, decorated with the @exec_ui.
                # Since we have used functools.wraps(), all our functions have the attribute __wrapped__
                # which points to the original function. What we need is the first line of the function
                # in the module file, because we want the functions to be sorted in the order they appear
                # in the source code file and not alphabetically.

                for name, func in sorted(funcs.items(), key=lambda x: x[1].__wrapped__.__code__.co_firstlineno):
                    # print(f"{func.__name__} -> {func.__wrapped__.__code__.co_firstlineno = }")
                    self._view.add_function_button(func)
            except ModuleNotFoundError as exc:
                rich.print(f"[red]{exc.__class__.__name__}: {exc}[/]")
                rich.print(f"Skipping '{mod}'...")



import argparse
import sys
from pathlib import Path

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication

from .config import load_config
from .control import Control
from .model import Model
from .view import View

HERE = Path(__file__).parent.resolve()

# EXAMPLE:
#   Use a function like this in your package '__init__' file and add it to the entry_points in the setup.py.
#   This function will load functions that are decorated with '@exec_ui' from all modules in 'camtest.contingency'
#   and dynamically create the GUI executor for them.
#
#     def contingency_ui():
#         cmd = ExternalCommand("gui-executor --module-path camtest.contingency", asynchronous=True)
#         cmd.start()
#
#   Replace 'camtest.contingency' by a package that contains the scripts you want to execute from the GUI.
#
#       entry_points={
#         "gui_scripts": [
#             "contingency_ui=camtest.contingency.__init__:contingency_ui",
#         ]
#     },


def main():

    parser = argparse.ArgumentParser(prog='gui-executor')
    parser.add_argument('--location', help='location of the Python modules and scripts')
    parser.add_argument('--module-path', help='module path of the Python modules and scripts')
    parser.add_argument('--config', help='a YAML file that configures the executor')
    parser.add_argument('--logo', help='path to logo PNG or SVG file')
    parser.add_argument('--app-name', help='the name of the GUI app, will go in the window title')

    args = parser.parse_args()

    # We have only implemented the --module-path option for now

    if args.module_path is None:
        print("You need to provide the --module-path option.")
        parser.print_help()
        return

    if args.config:
        load_config(args.config)

    app = QApplication([])
    app.setWindowIcon(QIcon(args.logo or str(HERE / "icons/tasks.svg")))

    view = View(args.app_name or "GUI Executor")
    model = Model(args.module_path)
    Control(view, model)

    view.show()

    return app.exec()


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import importlib
import inspect
import textwrap
import warnings
from enum import IntEnum
from functools import wraps
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple


class Kind(IntEnum):
    BUTTON = 0b00000001
    """Identifies a function to be called after a clicked event on a button in the GUI."""


class ArgumentKind(IntEnum):
    POSITIONAL_ONLY = 0
    POSITIONAL_OR_KEYWORD = 1
    VAR_POSITIONAL = 2
    KEYWORD_ONLY = 3
    VAR_KEYWORD = 4


class Argument:
    def __init__(self, name: str, kind: int, annotation: Any, default: Any):
        self.name = name
        self.kind: ArgumentKind = ArgumentKind(kind)
        self.annotation = annotation
        self.default = default


def exec_ui(
        kind: Kind = Kind.BUTTON,
        description: str = None,
        icon = None,
        input_request: Tuple[str, ...] = ("Continue [Y/n]", "Abort [Y/n]"),
        use_kernel: bool = False,
):
    """
    Decorates the function as an Exec UI function. We have different kinds of UI functions. By default,
    the function is decorated as a UI Button which will appear in the UI as a button to execute the function.

    Args:
        kind: identifies the function and what it can be used for [default = BUTTON]
        description: short function description intended to be used as tooltip or similar
        icon: an icon PNG to use for the button [size hints?]
        input_request: a tuple contain the string to detect when input is asked for
        use_kernel: use the Jupyter kernel when running this function

    Returns:
        The wrapper function object.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # you can put extra code to be run here, based on the arguments to exec_ui
            response = func(*args, **kwargs)
            # or here
            return response
        wrapper.__ui_kind__ = kind
        wrapper.__ui_description__ = description
        wrapper.__ui_file__ = func.__code__.co_filename
        wrapper.__ui_module__ = func.__module__
        wrapper.__ui_input_request__ = input_request
        wrapper.__ui_use_kernel__ = use_kernel
        return wrapper

    return decorator


def find_ui_button_functions(module_path: str) -> Dict[str, Callable]:
    """
    Returns a dictionary with function names as keys and the callable function as their value.
    The functions are intended to be used as UI button callable, i.e. a GUI can automatically
    identify these functions and assign them to a `clicked` action of a button.

    Args:
        module_path: string containing a fully qualified module name
    """
    return find_ui_functions(
        module_path,
        lambda x: x.__ui_kind__ & Kind.BUTTON
    )


def find_ui_functions(module_path: str, predicate: Callable = None) -> Dict[str, Callable]:
    """
    Returns a dictionary with function names as keys and the callable function as their value.
    The predicate is a function that returns True or False depending on some required conditions
    for the functions that are returned.

    Args:
        module_path: string containing a fully qualified module name
        predicate: condition to select and return the function
    """
    predicate = predicate if predicate is not None else lambda x: True
    mod = importlib.import_module(module_path)

    return {
        name: member
        for name, member in inspect.getmembers(mod)
        if inspect.isfunction(member) and hasattr(member, "__ui_kind__") and predicate(member)
    }


def find_modules(module_path: str) -> Dict[str, Any]:
    """
    Finds Python modules and scripts in the given module path (non recursively). The modules will not be
    imported, instead their module path will be returned. The idea is that the caller can decide which
    modules to import.

    Args:
        module_path: the module path where the Python modules and scripts are located

    Returns:
        A dictionary with module names as keys and their paths as values.
    """
    mod = importlib.import_module(module_path)

    if hasattr(mod, "__path__") and getattr(mod, "__file__", None) is None:
        warnings.warn(
            textwrap.dedent(f"""
                The module '{mod.__name__}' is a namespace package, i.e. a package without an '__init__.py' file.
                Please, properly define your module and add an '__init__.py' file. The file can be empty.
                Your package is located at {set(mod.__path__)}.
                """)
        )

        paths: List[str] = list(set(mod.__path__))
        location = Path(paths[0])
    else:
        location = Path(mod.__file__).parent

    if not location.is_dir():
        raise ValueError(f"Expected a folder, instead got {str(location)}")

    return {
        item.stem: f"{module_path}.{item.stem}"
        for item in location.glob("*.py")
        if item.name not in ["__init__.py"]
    }


# Why I use my own class Arguments instead of just inspect.Parameter?
# * because I don't want to be dependent on inspect.Parameter.empty in my apps
# * because Argument might get more info from the exec_ui decorator, like e.g. units
#   or a description of the argument
def get_arguments(func: Callable) -> Dict[str, Argument]:
    """
    Determines the signature of the function and returns a dictionary with keys the name of the arguments
    and values the Argument object for the arguments.

    Args:
        func: a function callable

    Returns:
        A dictionary with all arguments.
    """
    sig = inspect.signature(func)
    pars = sig.parameters
    return {
        k: Argument(
            k,
            int(v.kind),
            None if v.annotation == inspect.Parameter.empty else v.annotation,
            None if v.default == inspect.Parameter.empty else v.default
        )
        for k, v in pars.items()
    }



from __future__ import annotations

import os
import re
import sys
from copy import deepcopy
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple

from executor import ExternalCommand
from executor import ExternalCommandFailed

from gui_executor.kernel import MyKernel
from gui_executor.utils import expand_path
from gui_executor.utils import get_file_path


class CommandError(Exception):
    pass


class Command:
    def __init__(self, name: str,
                 path: Path = None, python_path: str = None, category: str = None, args: List[str] = None):
        self._name = name
        self._path = path
        self._category = category
        self._args = args or []
        self._python_path = python_path or ""

        self._parsed_args = None

    def execute(self, **kwargs) -> str | Command:
        return ""

    def can_execute(self) -> bool:
        return self._parsed_args is not None

    def set_python_path(self, python_path: str):
        # TODO:
        #    proper checking of the format of python_path
        self._python_path = python_path

    def get_python_path(self) -> str:
        return self._python_path

    def get_required_args(self) -> List[Tuple[str, str]]:
        """
        Returns a list with the required arguments and their expected types.
        """
        required_args = []
        for arg_name, arg_text in self._args:
            if m := re.search(r"<<([:\w]+)>>", arg_text):
                match = m[1]
                name, expected_type = match.split(':') if ':' in match else (match, None)
                required_args.append((name, expected_type))
        return required_args

    def parse_args(self, **kwargs):
        """
        Parses the arguments list and for each input request it will ask for input. When all input has been received
        the full argument list is constructed and returned as a string.

        Args:
            **kwargs: list of input arguments

        Returns:
            A string containing all arguments
        """
        parsed_args = ""
        for arg_name, arg_text in self._args:
            if m := re.search(r"<<([:\w]+)>>", arg_text):
                x, *_ = m[1].split(':')
                arg_value = kwargs[x] if x in kwargs else input(f"Enter a value for {x}: ")
            elif arg_text == "None":
                arg_value = ""
            else:
                arg_value = arg_text
            parsed_args += f"{arg_name} {arg_value} "

        self._parsed_args = parsed_args.strip()


class AppCommand(Command):
    def __init__(self, name: str, app_name: str):
        super().__init__(name)
        self._app_name = app_name


class ScriptCommand(Command):
    """This class represents a script command, i.e. a Python scripts which is executed as such."""
    def __init__(self, name: str, script_name: str,
                 env: dict = None,
                 path: Path = None, python_path: str = None, category: str = None, args: List[str] = None):
        super().__init__(name, path=path, python_path=python_path, category=category, args=args)
        self._script_name = script_name
        self._cmd: ExternalCommand | None = None
        self._env = env

    @staticmethod
    def from_config(config, name: str) -> ScriptCommand:
        from gui_executor.config import ExecutorConfiguration, ConfigError
        config: ExecutorConfiguration

        if "Scripts" not in config:
            raise ConfigError(f"No scripts defined in the configuration '{config.name}'")

        if name not in config.get_script_names():
            raise ConfigError(f"No script definition found for '{name}' in the configuration '{config.name}'.")

        python_path = config.get_python_path()
        env = config.get_environment()
        script: dict = config["Scripts"][name]
        script_name = script.get("script_name")
        # The path in a YAML file shall be absolute or relative to the YAML file location
        path = config.get_absolute_path(script.get("path"))
        category = script.get("category")
        args = script.get("args")

        return ScriptCommand(name, script_name,
                             env=env, path=path, python_path=python_path, category=category, args=args)

    def execute(self, capture: bool = True, asynchronous: bool = False) -> None:
        cmd_line = self.get_command_line()
        python_path = self.get_python_path()
        saved_env = None

        if self._env:
            saved_env = deepcopy(os.environ)
            os.environ.update(self._env)

        self._cmd = ExternalCommand(f"PYTHONPATH={python_path} {cmd_line}",
                                    capture=capture, capture_stderr=True, asynchronous=asynchronous)
        try:
            self._cmd.start()
        except ExternalCommandFailed as exc:
            raise CommandError(self._cmd.error_message) from exc
        finally:
            if saved_env is not None:
                os.environ = saved_env

    def is_running(self) -> bool:
        return self._cmd.is_running if self._cmd is not None else False

    def get_output(self) -> str:
        print(f"{self._cmd = }")
        return self._cmd.output if self._cmd is not None else ""

    def get_error(self) -> Optional[str]:
        return self._cmd.error_message

    def get_command_line(self) -> str:
        path = expand_path(self._path)
        if not path.exists():
            raise CommandError(f"The path '{self._path}' was expanded into '{path}' which doesn't exist.")

        filepath = path / self._script_name
        if not filepath.exists():
            raise CommandError(f"The generated filepath '{filepath}' doesn't exit for command script {self._name}")

        return f"{sys.executable} {filepath} {self._parsed_args}"


class SnippetCommand(Command):
    def __init__(self, name: str, code: str | List[str], env: dict = None, **kwargs):
        super().__init__(name, **kwargs)
        self._code = code
        self._env = env

        self._output = None
        self._error = None

    @staticmethod
    def from_config(config, name: str) -> SnippetCommand:
        from gui_executor.config import ExecutorConfiguration, ConfigError
        config: ExecutorConfiguration

        if "Snippets" not in config:
            raise ConfigError(f"No code snippets defined in the configuration '{config.name}'")

        if name not in config.get_snippet_names():
            raise ConfigError(f"No code snippet definition found for '{name}' in the configuration '{config.name}'.")

        python_path = config.get_python_path()
        env = config.get_environment()
        snippet: dict = config["Snippets"][name]
        # The path in a YAML file shall be absolute or relative to the YAML file location
        path = config.get_absolute_path(snippet.get("path"))
        category = snippet.get("category")
        args = snippet.get("args")

        if script_name := snippet.get("script_name"):
            with get_file_path(path, script_name).open(mode='r') as fd:
                code = fd.readlines()
        else:
            code = snippet.get("code").split('\n')

        print()
        print(f"{code = }")

        return SnippetCommand(name,
                              code=code, env=env, path=path, python_path=python_path, category=category, args=args)

    def execute(self, capture: bool = True, asynchronous: bool = False, kernel=None) -> None:
        saved_env = None

        if self._env:
            saved_env = deepcopy(os.environ)
            os.environ.update(self._env)

        kernel = kernel or MyKernel()
        code = '\n'.join(self._code)  # TODO: optionally might run each line separately?
        self._output = kernel.run_snippet(code)
        self._error = kernel.get_error()

        if saved_env is not None:
            os.environ = saved_env

    def get_output(self) -> str:
        return self._output

    def get_error(self) -> str:
        return self._error



"""
This module provides functionality to read the YAML configuration files and work with its content.
"""
from __future__ import annotations

import os
from pathlib import Path
from typing import Dict
from typing import List

import yaml
from rich.tree import Tree

from gui_executor.command import ScriptCommand
from gui_executor.command import SnippetCommand
from gui_executor.utils import walk_dict_tree


def load_config(filename: Path | str) -> ExecutorConfiguration:
    """Load the YAML config file from the given filename."""
    filename = Path(filename).expanduser().resolve()

    with filename.open(mode='r') as fd:
        config = yaml.safe_load(fd)

    config = ExecutorConfiguration(config, filename)
    config.check_config()

    return config


class ConfigError(Exception):
    pass


class ExecutorConfiguration:
    """
    This class provides easy access to the content of the Executor Configuration files.
    """
    def __init__(self, config: dict, filename: Path):
        """
        Args:
            config: an ExecutorConfiguration represented in a dictionary
            filename: full pathname of the configuration file
        """
        self._config: Dict = config
        self._filename: Path = filename
        self._name: str = filename.stem

    @property
    def name(self):
        return self._name

    def check_config(self) -> None:
        """
        Perform a basic format and content check on the config dictionary.

        Raises:
            A ConfigError when a problem is encountered.
        """
        if "Python Path" not in self._config:
            raise ConfigError(f"No 'Python Path' in the configuration file at {self._filename}")

    def get_script_names(self) -> List[str]:
        """
        Returns the names of the scripts that are defined in the configuration.
        An empty list is returned if no scripts are defined.
        """
        return list(self._config.get("Scripts", {}).keys())

    def get_app_names(self) -> List[str]:
        """
        Returns the names of the apps that are defined in the configuration.
        An empty list is returned if no apps are defined.
        """
        return list(self._config.get("Apps", {}).keys())

    def get_snippet_names(self) -> List[str]:
        """
        Returns the names of the snippets that are defined in the configuration.
        An empty list is returned if no snippets are defined.
        """
        return list(self._config.get("Snippets", {}).keys())

    def get_absolute_path(self, path: Path | str) -> Path:
        """
        Returns the absolute path for the given path. When a relative path is given, it is assumed
        this path is relative to the location of the config file.

        Args:
            path: a relative or absolute path name

        Returns:
            An absolute path (note that this absolute path may not exist, that is on the caller to check).
        """
        path = Path(path).expanduser()
        if path.is_absolute():
            return path

        config_path = self._filename.parent.resolve()
        return (config_path / path).resolve()

    def get_python_path(self):
        try:
            python_path = self._config["Python Path"]
            orig_python_path = os.environ.get("PYTHONPATH", "")
            prepend = ':'.join(python_path.get("prepend", ""))
            append = ':'.join(python_path.get("append", ""))
            python_path = f"{prepend}:{orig_python_path}:{append}"
        except KeyError:
            python_path = ""

        return python_path

    def get_environment(self):
        return self._config.get("Environment", {})

    def get_command_for_script(self, name: str) -> ScriptCommand:
        """
        Returns a ScriptCommand for the given script name.
        """
        return ScriptCommand.from_config(self, name)

    def get_command_for_snippet(self, name: str) -> SnippetCommand:
        """
        Returns a SnippetCommand for the given snippet name. Note that a snippet
        can also be a script.
        """
        return SnippetCommand.from_config(self, name)

    def __contains__(self, item):
        return item in self._config

    def __getitem__(self, item):
        return self._config[item]

    def __rich__(self):
        tree = Tree(self._name, guide_style="dim")
        walk_dict_tree(self._config, tree, text_style="dark grey")
        return tree



from typing import Any
from typing import Callable
from typing import Dict

from .exec import find_modules
from .exec import find_ui_button_functions


class Model:
    def __init__(self, module_path: str):
        self._module_path = module_path

    def reload_functions(self, mod):
        ...

    def get_ui_buttons_functions(self, mod: str) -> Dict[str, Callable]:
        return find_ui_button_functions(mod)

    def get_ui_modules(self) -> Dict[str, Any]:
        return find_modules(self._module_path)



from __future__ import annotations

import contextlib
import inspect
import os
import re
import sys
import textwrap
from io import StringIO
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple

from rich.text import Text
from rich.tree import Tree


def replace_environment_variable(input_string: str) -> str:
    """Returns the `input_string` with all occurrences of ENV['var'] expanded.

    >>> replace_environment_variable("ENV['HOME']/data/CSL")
    '/Users/rik/data/CSL'

    Args:
        input_string (str): the string to replace
    Returns:
        The input string with the ENV['var'] replaced, or None when the environment variable
        doesn't exist.
    """

    match = re.search(r"(.*)ENV\[['\"](\w+)['\"]\](.*)", input_string)
    if not match:
        return input_string
    pre_match, var, post_match = match[1], match[2], match[3]

    result = os.getenv(var, None)

    return pre_match + result + post_match if result else None


def walk_dict_tree(dictionary: dict, tree: Tree, text_style: str = "green"):
    """
    Walk recursively through the dictionary and add all nodes to the given tree.
    The tree is a Rich Tree object.
    """
    for k, v in dictionary.items():
        if isinstance(v, dict):
            branch = tree.add(f"[purple]{k}", style="", guide_style="dim")
            walk_dict_tree(v, branch, text_style=text_style)
        else:
            text = Text.assemble((str(k), "medium_purple1"), ": ", (str(v), text_style))
            tree.add(text)


def expand_path(path: Path | str) -> Path:
    """
    Returns the expanded absolute path.

    Args:
        path: a string representing a path segment or a Path

    Returns:
        An absolute path.
    """
    path = replace_environment_variable(str(path))
    path = Path(path).expanduser()

    return path.resolve()


def get_file_path(path: str | Path, name: str) -> Path:
    full_path = expand_path(path)
    if not full_path.exists():
        raise ValueError(f"The path '{full_path}' was expanded into '{path}' which doesn't exist.")

    filepath = full_path / name
    if not filepath.exists():
        raise ValueError(f"The generated filepath '{filepath}' doesn't exit for command script {name}")

    return filepath


def remove_ansi_escape(line):
    """
    Returns a new line where all ANSI escape sequences are removed.
    """
    ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    return ansi_escape.sub('', line)


def get_required_args(code: List | str) -> List[Tuple[str, str | None]]:
    """
    Returns a list of required arguments and their type.

    Args:
        code:

    Returns:

    """
    if isinstance(code, str):
        code = code.split('\n')

    required_args = []
    for line in code:
        if matches := re.findall(r"<<([:\w]+)>>", line):
            print(f"{matches = }")
            for match in matches:
                name, expected_type = match.split(':') if ':' in match else (match, None)
                required_args.append((name, expected_type))

    return required_args


def replace_required_args(code: List | str, args: List) -> List | str:

    code_lines = code.split('\n') if isinstance(code, str) else code

    new_code_lines = []
    for line in code_lines:
        if matches := re.findall(r"<<([:\w]+)>>", line):
            for match in matches:
                print(f"{match = }")
                name, expected_type = match.split(':') if ':' in match else (match, None)
                line = line.replace(f"<<{match}>>", f"****")
        new_code_lines.append(line)
    return new_code_lines


def var_exists(var_name: str):
    frame = inspect.currentframe()

    try:
        return var_name in frame.f_back.f_locals or var_name in frame.f_back.f_globals
    finally:
        del frame


@contextlib.contextmanager
def sys_path(path: Path | str):
    """Context manager that temporarily prepends the `sys.path` with the given argument."""
    import sys

    try:
        sys.path.insert(0, str(path))
        yield
    finally:
        sys.path.pop(0)


class Data(object):
    pass


@contextlib.contextmanager
def capture():
    stdout = sys.stdout
    stderr = sys.stderr
    out = StringIO()
    err = StringIO()
    sys.stdout = out
    sys.stderr = err
    data = Data()
    try:
        yield data
    finally:
        sys.stdout = stdout
        sys.stderr = stderr
        data.stdout = out.getvalue()
        data.stderr = err.getvalue()


def stringify_args(args):
    return ", ".join([repr(arg) for arg in args])


def stringify_kwargs(kwargs):
    return ", ".join([f"{k}={repr(v)}" for k, v in kwargs.items()])


def create_code_snippet(func: Callable, args: List, kwargs: Dict):

    return textwrap.dedent(
        f"""\
            from {func.__ui_module__} import {func.__name__}
            response = {func.__name__}({stringify_args(args)}{', ' if args else ''}{stringify_kwargs(kwargs)})
            if response is not None:
                print(response)
        """
    )



import logging
import queue
import textwrap

import jupyter_client.kernelspec
import rich
from executor import ExternalCommand
from executor import ExternalCommandFailed
from jupyter_client.manager import KernelManager
from rich import print

from gui_executor.utils import remove_ansi_escape

LOGGER = logging.getLogger("gui-executor.kernel")


class KernelError(Exception):
    pass


def decode_traceback(traceback: str) -> str:
    return remove_ansi_escape('\n'.join(traceback))


class MyKernel:

    def __init__(self, name: str = "python3", startup_timeout: int = 60):
        self._kernel = KernelManager(kernel_name=name)
        self._kernel.start_kernel()
        self._client = self._kernel.client()
        self._client.start_channels()
        try:
            self._client.wait_for_ready(timeout=startup_timeout)
        except RuntimeError:
            self._client.stop_channels()
            self._kernel.shutdown_kernel()
            raise
        self._error = None

    def is_alive(self) -> bool:
        return self._kernel.is_alive()

    def shutdown(self):
        self._kernel.shutdown_kernel(now=True)

    @staticmethod
    def get_kernel_specs():
        return jupyter_client.kernelspec.find_kernel_specs()

    def get_kernel_info(self):
        msg_id = self._client.kernel_info()
        # rich.print(f"get_info() -> {msg_id = }")
        reply = self._client.get_shell_msg()
        # rich.print("get_info() -> reply = ", reply)
        return reply

    def get_connection_file(self):
        return self._kernel.connection_file

    def run_snippet(self, snippet: str) -> str:

        # Flush the IOPub channel before executing the command. This is needed because another
        # client might be connected that has sent out messages on the pub channel. We want to
        # catch the output of the given command of course.

        self.flush()

        msg_id = self._client.execute(snippet)
        LOGGER.debug(f"{msg_id = }")

        reply = self._get_response(msg_id)
        LOGGER.debug(f"{type(reply) = }")
        LOGGER.debug(f"{reply = }")
        LOGGER.debug(f"{reply['content'] = }")

        if reply["content"]["status"] == "error":
            try:
                self._error = decode_traceback(reply["content"]["traceback"])
            except KeyError:
                self._error = "An error occurred, no traceback was provided."
        else:
            self._error = None

        output = self._get_output()
        # LOGGER.debug(f"{output = }")

        return output.strip()

    def flush(self):
        while True:
            try:
                _ = self._client.get_iopub_msg(timeout=0.2)
            except queue.Empty:
                break

    def get_error(self) -> str:
        return self._error

    def _get_response(self, msg_id: str = None) -> dict:
        return self._client.get_shell_msg(msg_id)

    def _get_output(self, timeout: float = 1.0) -> str:

        # When the execution state is "idle" it is complete

        io_msg_content = self._client.get_iopub_msg(timeout=timeout)['content']

        if 'execution_state' in io_msg_content and io_msg_content['execution_state'] == 'idle':
            return ""

        while True:
            last_io_msg_content = io_msg_content

            try:
                io_msg_content = self._client.get_iopub_msg(timeout=timeout)['content']
                if 'execution_state' in io_msg_content and io_msg_content['execution_state'] == 'idle':
                    break
            except queue.Empty:
                break

        return self._decode_io_msg_content(last_io_msg_content)

    @staticmethod
    def _decode_io_msg_content(content: dict) -> str:

        if 'data' in content:  # Indicates completed operation
            return content['data']['text/plain']
        elif 'name' in content and content['name'] == "stdout":  # indicates output
            return content['text']
        elif 'traceback' in content:  # Indicates an error
            return decode_traceback(content['traceback'])
        else:
            return ''

    def __del__(self):
        self._client.stop_channels()
        self._kernel.shutdown_kernel()


def do_test_my_kernel(name: str = "python3"):

    kernel = MyKernel(name=name)

    snippets = [
        "a=2",
        textwrap.dedent("""\
            a = 42
            b = 73
            c = a + b
            print(c)        
            """),
        'print(f"{a=}, {b=}, {c=}")',
        '1/0',  # should return a ZeroDivisionError
        'import sys; print(f"{sys.path = }")',
        'import pandas as pd',
        'df = pd.DataFrame(dict(A=[1,2,3], B=["one", "two", "three"]))',
        'df',
        'df.describe()',
        '!pip list -v'
    ]

    for snippet in snippets:
        if out := kernel.run_snippet(snippet):
            print(out)


def start_qtconsole(kernel: MyKernel, buffer_size: int = 5000, console_height: int = 42, console_width: int = 128):
    connection_file = kernel.get_connection_file()
    cmd_line = (f"jupyter qtconsole --ConsoleWidget.buffer_size={buffer_size} "
                f"--ConsoleWidget.console_height={console_height} "
                f"--ConsoleWidget.console_width={console_width} "
                f"--existing {connection_file}")

    cmd = ExternalCommand(
        f"{cmd_line}", capture=True, capture_stderr=True, asynchronous=True)
    try:
        cmd.start()
    except ExternalCommandFailed as exc:
        raise KernelError(cmd.error_message) from exc

    return cmd


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG)

    # kernel_name = "python3"
    kernel_name = "plato-common-egse"

    do_test_my_kernel(name=kernel_name)



import contextlib
import os
import pathlib
import sys
from shutil import rmtree

from setuptools import setup, Command

# Package meta-data.

NAME = "gui-executor"
PACKAGE_NAME = "gui_executor"
DESCRIPTION = "Execute Python code in an automatically generated GUI App."
URL = "https://github.com/rhuygen/gui-executor"
EMAIL = "rik.huygen@kuleuven.be"
AUTHOR = "Rik Huygen"
REQUIRES_PYTHON = '>=3.8.0'
VERSION = None

# The directory containing this file

HERE = pathlib.Path(__file__).parent

# The directory containing the source code

SRC = HERE / "src"

# The text of the README file

README = (HERE / "README.md").read_text()

# Load the package's __version__.py module as a dictionary.

about = {}
if VERSION is None:
    with open(os.path.join(SRC, PACKAGE_NAME, '__version__.py')) as f:
        exec(f.read(), about)
        VERSION = about['__version__']


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package to PyPI.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        with contextlib.suppress(OSError):
            self.status('Removing previous builds…')
            rmtree(os.path.join(HERE, 'dist'))
        self.status("Building Source and Wheel (universal) distribution…")
        os.system(f"{sys.executable} setup.py sdist bdist_wheel --universal")

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system(f"git tag v{VERSION}")
        os.system('git push --tags')

        sys.exit()


# This call to setup() does all the work

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Environment :: X11 Applications :: Qt",
        "Topic :: Utilities",
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    packages=["gui_executor", "gui_executor/icons"],
    package_dir={"": "src"},
    package_data={
        "": ["*.yaml", "*.png", "*.svg"],
    },
    include_package_data=False,
    install_requires=["rich", "PyQt5"],
    entry_points={
        "gui_scripts": [
            "gui-executor=gui_executor.__main__:main",
        ]
    },
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



from __future__ import annotations

import errno
import fcntl
import os
import select
import sys
import tempfile
import textwrap
import time
from functools import partial
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from PyQt5.QtCore import QObject
from PyQt5.QtCore import QRunnable
from PyQt5.QtCore import QSize
from PyQt5.QtCore import QThreadPool
from PyQt5.QtCore import QTimer
from PyQt5.QtCore import Qt
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtGui import QCloseEvent
from PyQt5.QtGui import QCursor
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtGui import QFont
from PyQt5.QtGui import QIcon
from PyQt5.QtGui import QIntValidator
from PyQt5.QtGui import QPainter
from PyQt5.QtGui import QTextCursor
from PyQt5.QtSvg import QSvgRenderer
from PyQt5.QtWidgets import QAction
from PyQt5.QtWidgets import QCheckBox
from PyQt5.QtWidgets import QComboBox
from PyQt5.QtWidgets import QFrame
from PyQt5.QtWidgets import QGridLayout
from PyQt5.QtWidgets import QGroupBox
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QMainWindow
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtWidgets import QTextEdit
from PyQt5.QtWidgets import QToolBar
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QWidget
from executor import ExternalCommand
from executor import ExternalCommandFailed
from rich.console import Console
from rich.text import Text

from .exec import Argument
from .exec import ArgumentKind
from .exec import get_arguments
from .kernel import MyKernel
from .kernel import start_qtconsole
from .utils import capture
from .utils import create_code_snippet
from .utils import stringify_args
from .utils import stringify_kwargs

HERE = Path(__file__).parent.resolve()


class VLine(QFrame):
    """Presents a simple Vertical Bar that can be used in e.g. the status bar."""

    def __init__(self):
        super().__init__()
        self.setFrameShape(QFrame.VLine | QFrame.Sunken)


class HLine(QFrame):
    """Presents a simple Horizontal Bar that can be used to separate widgets."""

    def __init__(self):
        super().__init__()
        self.setLineWidth(0)
        self.setMidLineWidth(1)
        self.setFrameShape(QFrame.HLine | QFrame.Sunken)


class FunctionThreadSignals(QObject):
    """
    Defines the signals available from a running function thread.

    Supported signals are:

    finished
        str: the function name
        bool: if the function was successfully executed, i.e. no exception was raised
    error
        str: Exception string
    data
        any object that was returned by the function
    """

    finished = pyqtSignal(str, bool)
    error = pyqtSignal(Exception)
    data = pyqtSignal(object)


# Helper function to add the O_NONBLOCK flag to a file descriptor
def make_async(fd):
    fcntl.fcntl(fd, fcntl.F_SETFL, fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NONBLOCK)


# Helper function to read some data from a file descriptor, ignoring EAGAIN errors
def read_async(fd):
    try:
        return fd.read()
    except IOError as exc:
        if exc.errno != errno.EAGAIN:
            raise exc
        else:
            return b''


class FunctionRunnable(QRunnable):
    def __init__(self, func: Callable, args: List, kwargs: Dict):
        super().__init__()
        self.signals = FunctionThreadSignals()
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self._check_for_input = False
        self._input_patterns = []

    def check_for_input(self, patterns: Tuple):
        self._check_for_input = True
        self._input_patterns = [pattern.rstrip() for pattern in patterns]

    def run(self):
        # We can in the future decide based on exec_ui arguments in how to run the function
        # self.run_in_current_interpreter()
        # self.run_in_kernel()
        # self.run_in_qprocess()
        self.run_in_external_command()

    def run_in_current_interpreter(self):
        # This runs the function within the current Python interpreter. This might be a security risk
        # if you allow to run functions that are not under your control.
        success = False
        self.signals.data.emit("-" * 20)
        self.signals.data.emit(
            f"Running function {self._func.__name__}({stringify_args(self._args)}{', ' if self._args else ''}"
            f"{stringify_kwargs(self._kwargs)})...")
        try:
            with capture() as out:
                response = self._func(*self._args, **self._kwargs)
            self.signals.data.emit(out.stdout)
            self.signals.data.emit(out.stderr)
            self.signals.data.emit(response)
            success = True
        except Exception as exc:
            self.signals.data.emit(out.stdout)
            self.signals.data.emit(out.stderr)
            self.signals.error.emit(exc)
            success = False
        finally:
            self.signals.finished.emit(self._func.__name__, success)

    def run_in_kernel(self):
        ...

    def run_in_qprocess(self):
        ...

    def run_in_external_command(self):
        tmp = tempfile.NamedTemporaryFile(mode='w', delete=False)
        tmp.write(create_code_snippet(self._func, self._args, self._kwargs))
        tmp.close()

        self.signals.data.emit("-" * 20)

        options = dict(capture=True, capture_stderr=True, asynchronous=True, buffered=False, input=True)
        try:
            # We could actually just use SubProcess here to with the correct settings
            with ExternalCommand(f"{sys.executable} {tmp.name}", **options) as cmd:
                make_async(cmd.stdout)
                make_async(cmd.stderr)

                while True:
                    # Wait for data to become available
                    out, *_ = select.select([cmd.stdout, cmd.stderr], [], [])

                    # Try reading some data from each
                    line_out = read_async(cmd.stdout)
                    line_err = read_async(cmd.stderr)

                    if line_out:
                        self.signals.data.emit(line := line_out.decode(cmd.encoding).rstrip())
                        # Try to detect when the process is requesting input.
                        # TODO: request input from user through QLineEdit field...
                        if self._check_for_input and any(pattern in line for pattern in self._input_patterns):
                            time.sleep(1.0)
                            cmd.subprocess.stdin.write(b'Y\n')
                    if line_err:
                        self.signals.data.emit(line := line_err.decode(cmd.encoding).rstrip())

                    if cmd.subprocess.poll() is not None:
                        break

                # Previous attempts that didn't work as expected. Need to document this....

                # while cmd.is_running:
                #     if cmd.stdout.readable():
                #         out = cmd.stdout.readline().decode(cmd.encoding)
                #         self.signals.data.emit(out.rstrip())
                #     if cmd.stderr.readable():
                #         err = cmd.stderr.readline().decode(cmd.encoding)
                #         self.signals.data.emit(err.rstrip())

                # for out, err in itertools.zip_longest(
                #         # iter(lambda: cmd.stdout.readline().decode(cmd.encoding), u''),
                #         # iter(lambda: cmd.stderr.readline().decode(cmd.encoding), u'')
                #         iter(lambda: cmd.stdout.readline().decode(cmd.encoding), u''),
                #         iter(lambda: cmd.stderr.readline().decode(cmd.encoding), u'')
                # ):
                #     out and self.signals.data.emit(out.rstrip())
                #     err and self.signals.data.emit(err.rstrip())

            cmd.wait()

        except ExternalCommandFailed as exc:
            # self._console_panel.append(cmd.error_message)
            # This error message is also available in the decoded_stderr.
            self.signals.error.emit(exc)

        # if out := cmd.decoded_stdout:
        #     self._console_panel.append(out)
        if err := cmd.decoded_stderr:
            self.signals.data.emit(err)

        if cmd.is_finished:
            if cmd.failed:
                self.signals.finished.emit(self._func.__name__, False)
            else:
                self.signals.finished.emit(self._func.__name__, True)

        # print(f"{tmp.name = }")
        os.unlink(tmp.name)

    def start(self):
        QThreadPool.globalInstance().start(self)


class ConsoleOutput(QTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setReadOnly(True)
        self.setLineWrapMode(QTextEdit.NoWrap)
        self.insertPlainText("")
        self.setMinimumSize(600, 300)
        monospaced_font = QFont("Courier New")
        monospaced_font.setStyleHint(QFont.Monospace)
        monospaced_font.setPointSize(14)  # TODO: should be a setting
        self.setFont(monospaced_font)

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.__contextMenu)

    @pyqtSlot(str)
    def append(self, text):
        self.moveCursor(QTextCursor.End)

        console = Console(record=True, force_terminal=False, force_jupyter=False)
        with console.capture():
            console.print(text)

        exported_html = console.export_html(
            inline_styles=True,
            code_format="<pre style=\"font-family:Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace\">{code}\n</pre>"
        )

        self.insertHtml(exported_html)

        sb = self.verticalScrollBar()
        sb.setValue(sb.maximum())

    def __contextMenu(self):
        self._normalMenu = self.createStandardContextMenu()
        self._addCustomMenuItems(self._normalMenu)
        self._normalMenu.exec_(QCursor.pos())

    def _addCustomMenuItems(self, menu):
        menu.addSeparator()
        menu.addAction(u'Clear', self.clear)


class IconLabel(QLabel):

    IconSize = QSize(16, 16)

    def __init__(self, icon_path: Path | str, size: QSize = IconSize):
        super().__init__()

        self.icon_path = str(icon_path)
        self.setFixedSize(size)

    def paintEvent(self, *args, **kwargs):

        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setPen(Qt.NoPen)

        renderer = QSvgRenderer()
        renderer.load(self.icon_path)
        renderer.render(painter)

        painter.end()


class DynamicButton(QWidget):

    IconSize = QSize(30, 30)
    HorizontalSpacing = 2

    def __init__(self, label: str, func: Callable,
                 icon_path: Path | str = None, final_stretch=True, size: QSize = IconSize):
        super().__init__()
        self._function = func
        self._label = label
        self._icon = None

        self.icon_path = str(icon_path or HERE / "icons/023-evaluate.svg")

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        label_icon = IconLabel(icon_path=self.icon_path, size=size)
        label_text = QLabel(label)

        layout.addWidget(label_icon)
        layout.addSpacing(self.HorizontalSpacing)
        layout.addWidget(label_text)

        if final_stretch:
            layout.addStretch()

        self.setToolTip(func.__doc__)

    @property
    def function(self) -> Callable:
        return self._function

    @property
    def module_name(self) -> str:
        """Returns the name of the module where the function resides."""
        return self._function.__ui_module__

    @property
    def label(self) -> str:
        return self._label

    def __repr__(self):
        return f"DynamicButton(\"{self.label}\", {self.function})"


class ArgumentsPanel(QGroupBox):
    def __init__(self, button: DynamicButton, ui_args: Dict[str, Argument]):
        super().__init__(button.label)

        self._button = button
        self._ui_args = ui_args
        self._args_fields = {}
        self._kwargs_fields = {}

        # The arguments panel is a Widget with an input text field for each of the arguments.
        # The text field is pre-filled with the default value if available.

        vbox = QVBoxLayout()
        grid = QGridLayout()

        for idx, (name, arg) in enumerate(ui_args.items()):
            if arg.annotation is bool:
                input_field = QCheckBox("")
                input_field.setCheckState(Qt.Checked if arg.default else Qt.Unchecked)
            else:
                input_field = QLineEdit()
                input_field.setObjectName(name)
                input_field.setPlaceholderText(str(arg.default) if arg.default else "")
                if arg.annotation is not None:
                    input_field.setToolTip(f"The expected type is {arg.annotation.__name__}.")
                else:
                    input_field.setToolTip("No type has been specified..")
                if arg.annotation is int:
                    input_field.setValidator(QIntValidator())
                elif arg.annotation is float:
                    input_field.setValidator(QDoubleValidator())

            if arg.kind == ArgumentKind.POSITIONAL_ONLY:
                self._args_fields[name] = input_field
            elif arg.kind in [ArgumentKind.POSITIONAL_OR_KEYWORD, ArgumentKind.KEYWORD_ONLY]:
                self._kwargs_fields[name] = input_field
            else:
                print("ERROR: Only POSITIONAL_ONLY, POSITIONAL_OR_KEYWORD, and KEYWORD_ONLY arguments are supported!")

            label = QLabel(name)
            type_hint = QLabel(f"[{arg.annotation.__name__}]" if arg.annotation is not None else None)
            type_hint.setStyleSheet("color: gray")

            # Stretch the middle column of the grid. That is needed when there is only one argument and it's a bool
            # i.e. a CheckBox. If we do not stretch, the checkbox will be centered.
            grid.setColumnStretch(1, 1)
            grid.addWidget(label, idx, 0)
            grid.addWidget(input_field, idx, 1)
            grid.addWidget(type_hint, idx, 2)

        vbox.addLayout(grid)

        hbox = QHBoxLayout()
        self.kernel_checkbox = QCheckBox("run in kernel")
        self.kernel_checkbox.setCheckState(Qt.Checked if self.function.__ui_use_kernel__ else Qt.Unchecked)
        self.run_button = QPushButton("run")
        hbox.addWidget(self.kernel_checkbox)
        hbox.addStretch()
        hbox.addWidget(self.run_button)

        vbox.addLayout(hbox)

        self.setLayout(vbox)

    @property
    def function(self):
        return self._button.function

    @property
    def args(self):
        return [
            self._cast_arg(name, field)
            for name, field in self._args_fields.items()
        ]

    @property
    def kwargs(self):
        return {
            name: self._cast_arg(name, field)
            for name, field in self._kwargs_fields.items()
        }

    @property
    def use_kernel(self):
        return self.kernel_checkbox.checkState() == Qt.Checked

    def _cast_arg(self, name: str, field: QLineEdit | QCheckBox):
        arg = self._ui_args[name]

        if arg.annotation is bool:
            return field.checkState() == Qt.Checked

        value = field.displayText() or field.placeholderText()

        try:
            return arg.annotation(value)
        except (ValueError, TypeError):
            return value


class FunctionButtonsPanel(QWidget):
    def __init__(self):
        super().__init__()

        self.n_cols = 4  # This must be a setting or configuration option

        # The modules are arranged in a vertical layout and each of the functions in that module is arranged in a
        # horizontal layout. Modules are added when a new button is added for a not yet existing module.

        self.modules: Dict[str, QGridLayout] = {}
        self.buttons: Dict[str, int] = {}
        self.module_layout = QVBoxLayout()

        self.setLayout(self.module_layout)

    def add_button(self, button: DynamicButton):
        module_name = button.module_name
        if module_name not in self.modules:
            grid = QGridLayout()
            # Make sure all columns have equal width
            for idx in range(self.n_cols):
                grid.setColumnStretch(idx, 1)
            gbox = QGroupBox(module_name.rsplit(".", 1)[-1])
            gbox.setLayout(grid)
            gbox.setStyleSheet(textwrap.dedent("""
                QGroupBox
                {
                    font-size: 14px;
                    font-weight: light;
                    color: grey;
                }
            """))
            self.module_layout.addWidget(gbox)
            self.modules[module_name] = grid
            self.buttons[module_name] = 0

        self.buttons[module_name] += 1
        button_count = self.buttons[module_name]

        row = (button_count - 1) // self.n_cols
        col = (button_count - 1) % self.n_cols
        self.modules[module_name].addWidget(button, row, col)


class KernelPanel(QWidget):
    def __init__(self):
        super().__init__()

        hbox = QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)

        kernel_specs = list(MyKernel.get_kernel_specs())
        try:
            idx = kernel_specs.index("python3")
        except ValueError:
            idx = 0
        self.kernel_list = QComboBox()
        self.kernel_list.addItems(list(kernel_specs))
        self.kernel_list.setCurrentIndex(idx)

        hbox.addStretch(1)
        hbox.addWidget(QLabel("available kernels"))
        hbox.addWidget(self.kernel_list)

        self.setLayout(hbox)

    @property
    def selected_kernel(self) -> str:
        return self.kernel_list.currentText()


class View(QMainWindow):
    def __init__(self, app_name: str = None):
        super().__init__()

        self._qt_console: Optional[ExternalCommand] = None
        self._kernel: Optional[MyKernel] = None
        self._buttons = []
        self.function_thread: FunctionRunnable

        self.setWindowTitle(app_name or "GUI Executor")

        # self.setGeometry(300, 300, 500, 200)

        # The main frame in which all the other frames are located, the outer Application frame

        self.app_frame = QFrame()
        self.app_frame.setObjectName("AppFrame")

        # We don't want this QFrame to shrink below 500 pixels, therefore set a minimum horizontal size
        # and set the policy such that it can still expand from this minimum size. This will be used
        # when we use adjustSize after replacing the arguments panel.

        self.app_frame.setMinimumSize(500, 0)  # TODO: should be a setting
        sp = self.app_frame.sizePolicy()
        sp.setHorizontalPolicy(QSizePolicy.MinimumExpanding)
        self.app_frame.setSizePolicy(sp)

        self._layout_panels = QVBoxLayout()
        self._layout_buttons = FunctionButtonsPanel()

        self._layout_panels.addWidget(self._layout_buttons)
        self._layout_panels.addWidget(HLine())
        self._current_args_panel: QWidget = QWidget()
        self._current_args_panel.hide()
        self._layout_panels.addWidget(self._current_args_panel)
        self._console_panel = ConsoleOutput()
        self._layout_panels.addWidget(HLine())
        self._layout_panels.addWidget(self._console_panel)

        self.app_frame.setLayout(self._layout_panels)

        self.setCentralWidget(self.app_frame)

        QTimer.singleShot(500, self.start_kernel)

        self._rich_console = Console(force_terminal=False, force_jupyter=False)

        self._toolbar = QToolBar()
        self._toolbar.setIconSize(QSize(40, 40))
        self.addToolBar(self._toolbar)

        # Add a button to the toolbar to restart the kernel

        kernel_button = QAction(QIcon(str(HERE / "icons/reload-kernel.svg")), "Restart the Jupyter kernel", self)
        kernel_button.setStatusTip("Restart the Jupyter kernel")
        kernel_button.triggered.connect(partial(self.start_kernel, False))
        kernel_button.setCheckable(False)
        self._toolbar.addAction(kernel_button)

        # Add a button to the toolbar to start the qtconsole

        qtconsole_button = QAction(QIcon(str(HERE / "icons/command.svg")), "Start Qt Console", self)
        qtconsole_button.setStatusTip("Start the QT Console")
        qtconsole_button.triggered.connect(self.start_qt_console)
        qtconsole_button.setCheckable(False)
        self._toolbar.addAction(qtconsole_button)
        self.kernel_panel = KernelPanel()
        self._toolbar.addWidget(self.kernel_panel)

    def closeEvent(self, event: QCloseEvent) -> None:
        if self._kernel:
            self._kernel.shutdown()

    def start_kernel(self, force: bool = False) -> MyKernel:

        # Starting the kernel will need a proper PYTHONPATH for importing the packages

        if force or self._kernel is None:
            self._start_new_kernel()
        else:
            button = QMessageBox.question(
                self,
                "Restart Jupyter kernel", "A kernel is running, should a new kernel be started?"
            )
            if button == QMessageBox.Yes:
                self._console_panel.append('-' * 50)
                self._start_new_kernel()
        return self._kernel

    def _start_new_kernel(self):
        name = self.kernel_panel.selected_kernel
        self._kernel = MyKernel(name)
        self._console_panel.append(f"New kernel '{name}' started...")
        info = self._kernel.get_kernel_info()
        if 'banner' in info['content']:
            self._console_panel.append(info['content']['banner'])

    def start_qt_console(self):
        if self._qt_console is not None and self._qt_console.is_running:
            dialog = QMessageBox.information(self, "Qt Console", "There is already a Qt Console running.")
        else:
            self._qt_console = start_qtconsole(self._kernel or self.start_kernel())

    def run_function(self, func: Callable, args: List, kwargs: Dict, use_kernel: bool):
        if use_kernel:
            # remember kernel setting for this function
            func.__ui_use_kernel__ = True
            self.run_function_in_kernel(func, args, kwargs)
        else:
            self.run_function_in_thread(func, args, kwargs)

    def run_function_in_thread(self, func: Callable, args: List, kwargs: Dict):

        # TODO:
        #  * disable run button (should be activate again in function_complete?)

        self.function_thread = worker = FunctionRunnable(func, args, kwargs)
        self.function_thread.check_for_input(func.__ui_input_request__)
        self.function_thread.start()

        worker.signals.data.connect(self.function_output)
        worker.signals.finished.connect(self.function_complete)
        worker.signals.error.connect(self.function_error)

    def run_function_in_kernel(self, func: Callable, args: List, kwargs: Dict):
        self._kernel = self._kernel or MyKernel()

        self.function_output("-" * 20)

        snippet = create_code_snippet(func, args, kwargs)

        if response := self._kernel.run_snippet(snippet):
            self.function_output(response)

        self.function_complete(func.__name__, True)

    def add_function_button(self, func: Callable):

        button = DynamicButton(func.__name__, func)
        button.mouseReleaseEvent = partial(self.the_button_was_clicked, button)
        # button.clicked.connect(partial(self.the_button_was_clicked, button))

        self._buttons.append(button)
        self._layout_buttons.add_button(button)

    def the_button_was_clicked(self, button: DynamicButton, *args, **kwargs):

        # TODO
        #   * This should be done from the control or model and probably in the background?
        #   * Add ArgumentsPanel in a tabbed widget? When should it be removed from the tabbed widget? ...

        ui_args = get_arguments(button.function)

        args_panel = ArgumentsPanel(button, ui_args)
        args_panel.run_button.clicked.connect(
            lambda checked: self.run_function(args_panel.function, args_panel.args, args_panel.kwargs, args_panel.use_kernel)
        )

        if self._current_args_panel:
            self._layout_panels.replaceWidget(self._current_args_panel, args_panel)
            self._current_args_panel.setParent(None)
        else:
            self._layout_panels.addWidget(args_panel)

        self._current_args_panel = args_panel
        # self.app_frame.adjustSize()
        # self.adjustSize()

    @pyqtSlot(object)
    def function_output(self, data: object):
        self._console_panel.append(str(data))

    @pyqtSlot(str, bool)
    def function_complete(self, name: str, success: bool):
        if success:
            self._console_panel.append(f"function '{name}' execution finished.")
        else:
            self._console_panel.append(f"function '{name}' raised an Exception.")

    @pyqtSlot(Exception)
    def function_error(self, msg: Exception):
        text = Text.styled(f"{msg.__class__.__name__}: {msg}", style="bold red")
        self._console_panel.append(msg)
